package uk.co.hcsoftware.image;

import com.sun.media.jai.codec.FileSeekableStream;
import ns.TaskTimer;
import org.apache.commons.cli.*;

import javax.media.jai.Interpolation;
import javax.media.jai.JAI;
import javax.media.jai.PlanarImage;
import java.awt.image.renderable.ParameterBlock;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;


/**
 * scales a file to a particular output size (specify maxW, maxH) optionally can preserve aspect ratio
 */
public class ImageConverter {
    public static final Option verbose = new Option("v","verbose",false,"print out args and other info");

    public static final Option maxW = new Option("w","maxwidth",true,"(required) max width of output image");
    public static final Option maxH = new Option("h","maxheight",true,"(required) max height of output image");

    public static final Option outputFormat = new Option("o","outputformat",true,"format for output image (default=jpeg) see javax.media.jai.operator.FileStoreDescriptor for supported formats");
    public static final Option preserveAspect = new Option("p","preserveaspect",true,"perserve aspect ratio (default=true)");

    private static final Options availableOptions = new Options();
    static{
        maxW.setRequired(true);
        maxH.setRequired(true);
        try {
            Field[] fields = ImageConverter.class.getDeclaredFields();
            for (int i = 0; i < fields.length; i++) {
                Field field = fields[i];
                if(field.getType()==Option.class){ //all fields that are Option s
                    availableOptions.addOption((Option)field.get(null));
                }
            }
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        }
        verbose.setRequired(false);
    }
   // static final OptionBuilder ob = new OptionBuilder();

    private static CommandLine cmd;
    public static void main(String[] args) {
        try {
            cmd = new GnuParser().parse(availableOptions,args);


            if(cmd.hasOption(verbose.getOpt())){
               Option[] opts = cmd.getOptions();
               StringBuffer sb = new StringBuffer();
               for (int i = 0; i < opts.length; i++) {
                   Option opt = opts[i];
                   sb.append(opt.getOpt()+"="+opt.getValue());
                   if(i+1<opts.length) sb.append(" ");
               }
               System.out.println("options: "+sb.toString());
               System.out.println("args: "+cmd.getArgList());
           }
           if(cmd.getArgs().length!=2 || "help".equals(args[0])){
               System.out.println("args="+cmd.getArgList());
               usage();
               return;
           }
            String inFile=cmd.getArgs()[0];
            String outFile=cmd.getArgs()[1];
            processFile(new File(inFile), outFile);

            return;
        } catch(MissingOptionException e){
            System.err.println("Missing option(s): "+e.getMessage());
        } catch (ParseException e) {
            e.printStackTrace();
        }
        usage();
    }

    private static void processFile(File inFile, String outFile) {
        /*
        * Create an input stream from the specified file name
        * to be used with the file decoding operator.
        */
        FileSeekableStream stream = null;
        try {
            stream = new FileSeekableStream(inFile);
        } catch (IOException e) {
            e.printStackTrace();
            System.exit(0);
        }
        final TaskTimer tt = new TaskTimer();

        /* Create an operator to decode the image file. */
        PlanarImage pi = JAI.create("stream", stream);
        final boolean preserveAspectRatio = "true".equals(cmd.getOptionValue(preserveAspect.getOpt(),"true"));
        pi= new ImageConverter().scale(pi
                ,Integer.parseInt(cmd.getOptionValue(maxW.getOpt()))
                ,Integer.parseInt(cmd.getOptionValue(maxH.getOpt()))
                , preserveAspectRatio);
        ParameterBlock pb = new ParameterBlock();
        pb.addSource(pi);
        pb.add(outFile);
        final String of = cmd.getOptionValue(outputFormat.getOpt(),"jpeg");
        pb.add(of);
        pb.add(null);
        //  pb.add(true);
        JAI.create("filestore", pb, null);
        System.out.println("wrote file= " + outFile+ " took "+tt.stopString());
    }

    private static void usage(){
      //  System.out.println(Arrays.asList(args));
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp( "ImageConverter options inputimagefile outputimagefile ", availableOptions );
    }
   /*  private static void usage(){
        HelpFormatter formatter = new HelpFormatter();
         final PrintWriter printWriter = new PrintWriter(System.out);
         formatter.printUsage(printWriter,80,ImageConverter.class.getName(),availableOptions);
         printWriter.flush();
         printWriter.close();
    }*/


    private PlanarImage scale(PlanarImage pi, int maxOutputWidth, int maxOutputHeight, boolean preserveAspectRatio){
        int startWidth=pi.getWidth();
        int startHeight=pi.getHeight();

        System.out.println("input w="+startWidth+" h="+startHeight);
        float xScaleFactor=(float)maxOutputWidth/(float)startWidth;
        float yScaleFactor=(float)maxOutputHeight/(float)startHeight;
        //preserve the aspect ratio.. i.e. make sure x and y are the same

        if(preserveAspectRatio && xScaleFactor!=yScaleFactor){
            //whichever is the smaller
            xScaleFactor=Math.min(xScaleFactor,yScaleFactor);
            yScaleFactor=xScaleFactor;
        }
        System.out.println("xScaleFactor = " + xScaleFactor);
        System.out.println("yScaleFactor = " + yScaleFactor);

        /*
        * Create a standard bilinear interpolation object to be
        * used with the "scale" operator.
        */
        Interpolation interp = Interpolation.getInstance(Interpolation.INTERP_BILINEAR);

        /**
         * Stores the required input source and parameters in a
         * ParameterBlock to be sent to the operation registry,
         * and eventually to the "scale" operator.
         */
        ParameterBlock params = new ParameterBlock();
        params.addSource(pi);
        params.add(xScaleFactor);         // x scale factor
        params.add(yScaleFactor);         // y scale factor
        params.add(0.0F);         // x translate
        params.add(0.0F);         // y translate
        params.add(interp);       // interpolation method

        /* Create an operator to scale image1. */
        return JAI.create("scale", params);

    }

}
