package uk.co.hcsoftware.images.gui;

import ns.TaskTimer;
import uk.co.hcsoftware.images.UsageException;
import uk.co.hcsoftware.images.tools.ImageOperations;
import uk.co.hcsoftware.yago.ArgReader;
import uk.co.hcsoftware.yago.InvalidArgException;
import uk.co.hcsoftware.yago.Option;
import uk.co.hcsoftware.yago.parsers.GnuParser;

import javax.swing.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.KeyEvent;
import java.awt.image.ImageObserver;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.prefs.Preferences;

/**
 * image viewer app for flicking through images/thumbnails quickly.
 *
 * Tried windows native libs and didn't seem any faster to load/scale/display..
 * was taking ~937ms to display ~1.5mb png files
 * java version was taking ~1s
 *   names:
 *
 * from http://java.sun.com/j2se/1.5.0/docs/guide/2d/new_features.html
 *   -Dsun.java2d.opengl=true
 *   -Dsun.java2d.opengl=True
 *
 */
public class ImageMonkey2 extends JFrame implements ImageObserver{
    private static final Preferences prefs = Preferences.userNodeForPackage(ImageMonkey2.class);
    private static final Logger log = Logger.getLogger(ImageMonkey2.class.getName());
    private ImageObserver nullObserver=new ImageObserver() {
        public boolean imageUpdate(Image img, int infoflags, int x, int y, int width, int height) {
            return true;
        }
    };

    static enum ScaleType{
        //    SCALE_DEFAULT, SCALE_FAST, SCALE_SMOOTH, SCALE_REPLICATE, SCALE_AREA_AVERAGING
        deflt("default",Image.SCALE_DEFAULT)
        ,fast("fast",Image.SCALE_FAST)
        ,smooth("smooth",Image.SCALE_SMOOTH)
        ,replicate("replicate",Image.SCALE_REPLICATE)
        ,areaaveraging("area_avg",Image.SCALE_AREA_AVERAGING);
        private String name;
        private int constVal;

        ScaleType(String name, int constVal) {
            this.name = name;
            this.constVal = constVal;
        }

        public String toString() {
            return name;
        }

        public String getName() {
            return name;
        }

        public int getConstVal() {
            return constVal;
        }
        public static String getValuesPrintout(){
            StringBuilder sb = new StringBuilder();
            for(ScaleType st:values()){
                sb.append(st+" ");
            }
            return sb.toString();
        }
    }
    static final Option<ScaleType> scaleType = new Option<ScaleType>("scaletype"
            ,false
            ,new ArgReader<ScaleType>() {
                public ScaleType readArg(String argValue) throws InvalidArgException {
                    for(ScaleType st: ScaleType.values()){
                        if(st.getName().equals(argValue)) return st;
                    }
                    throw new InvalidArgException("must be one of: "+ScaleType.getValuesPrintout()+" was="+argValue,argValue);
                }
            }
            ,"Scaling type, values correspond to the constants in java.awt.Image and are: "+ScaleType.getValuesPrintout());



    public static void main(String args[]) {
        GnuParser gp=new GnuParser(scaleType);
        gp.parseOrDie("ImageMonkey",args);
        try {
            List<String> arguments =gp.getExtraArgs();
            if(arguments.size()==0 ){
                throw new UsageException("0 args");
            }
            if(Arrays.asList(new String[]{"--help","help","h","-h"}).contains(arguments.get(0))){
                throw new UsageException("invalid args");
            }
            List<File> imageFiles = new ArrayList<File>(arguments.size());
            for(String arg:arguments){
                imageFiles.add(new File(arg));
            }
            String s = System.getProperty("sun.java2d.opengl");
            if(s==null){
                System.setProperty("sun.java2d.opengl","True");
            }
            new ImageMonkey2(imageFiles);
            return;
        } catch(UsageException e){
            ;
        }catch (Exception e) {
            e.printStackTrace();
        }
        String s = gp.getUsage(ImageMonkey2.class.getName()+" imagefile1 [imagefilen..]",getLongDescription());
        System.out.println(s);
    }

    private static String getLongDescription() {
        return //ImageMonkey.class.getName()+" imagefile1 [imagefilen..]" +
                "\n\tdisplays a window with the first image." +
                "\n\timages are scaled to fit the window" +

                "\n\n Controls:"+
                "\n\tPgDn,n,SPACE:\tnext image" +
                "\n\tPgUp,p,DEL:\t\tprev image" +
                "\n\tHOME:\t\t\tfirst image" +
                "\n\tEND:\t\t\tlast image" +
                "\n"+
                "\n\tr:\t\t\treload current image" +
                "\n"+
                // "(textfiles holding 1 filename per line, created in workingdir, called imagenames_${keyname})"+
                "\n\tF1-F5:\t\t\tappend filename to quickfile" +
                "\n\tSHIFT+[F1-F5]:\tremove curr imagename from quickfile" +
                "\n\tCTRL+[F1-F5]:\tmove curr imagename to quickfile"
                ;
    }

    final List<File> imageFiles;

    public ImageMonkey2(final List<File> imageFiles) {
        super("imageviewer");

        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        this.imageFiles=imageFiles;

        setupKeys(imageFiles);

        //restore size
        int h=prefs.getInt("frameheight",600);
        int w=prefs.getInt("framewidth",800);
        log.finer("restore framesize prop="+w+" h="+h);
        if(w<10 || h<10){
            log.warning("unsensible framesizes, using defaults prop="+w+" h="+h);
            w=800; h=600;
        }
        setSize(w,h);

        addComponentListener(new ComponentAdapter() {
            public void componentResized(ComponentEvent e) {
                final int h = getHeight();
                final int w = getWidth();
                log.finer("saving framesize prop="+w+" h="+h);
                prefs.putInt("frameheight",h);
                prefs.putInt("framewidth",w);
                log.info("component resized: repainting");
                repaint();
                cache.clear();
            }
        });

        setVisible(true);
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                showImage(0);
            }
        });
    }
    final static Font font = new Font("SanSerif",Font.PLAIN,20);
    public void paint(Graphics g) {
        drawBlankScreen(g);
        if (gotImage) {
            g.drawImage(image, 0, 0,this);
        }
    }

    private void drawBlankScreen(Graphics g) {
        g.setColor(Color.WHITE);
        g.fillRect(0,0,getWidth(),getHeight());
    }

    private void setupKeys(final List<File> imageFiles) {
        final InputMap inputMap = ((JPanel)getContentPane()).getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW);
        final ActionMap actionMap = ((JPanel)getContentPane()).getActionMap();

        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_DOWN,0),"next");
        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_N,0),"next");
        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_SPACE,0),"next");
        actionMap.put("next",new AbstractAction() {
            public void actionPerformed(ActionEvent event) {
                showImage(index+1>=imageFiles.size()?0:index+1);
            }
        });
        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_UP,0),"previous");
        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_DELETE,0),"previous");
        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_P,0),"previous");
        actionMap.put("previous",new AbstractAction() {
            public void actionPerformed(ActionEvent event) {
                showImage(index==0?imageFiles.size()-1:index-1);
            }
        });
        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_HOME,0),"first");
        actionMap.put("first",new AbstractAction() {
            public void actionPerformed(ActionEvent event) {
                showImage(0);
            }
        });
        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_END,0),"last");
        actionMap.put("last",new AbstractAction() {
            public void actionPerformed(ActionEvent event) {
                showImage(imageFiles.size()-1);
            }
        });
        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_R,0),"reload");
        actionMap.put("reload",new AbstractAction() {
            public void actionPerformed(ActionEvent event) {
                showImage(index);
            }
        });

        setupQuickfilesKeys(inputMap, actionMap, imageFiles);

        /*inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_MINUS,0),"zoomout");
        actionMap.put("zoomout",new AbstractAction() {
        public void actionPerformed(ActionEvent event) {
        log.log(Level.INFO,"zoomout");
        showImage(index, zoomFactor-1);
        }
        });*/
        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_O,0),"noscale");
        actionMap.put("noscale",new AbstractAction() {
            public void actionPerformed(ActionEvent event) {
                showImage(index);
            }
        });


        //image edit actions
        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_R,KeyEvent.CTRL_MASK),"rotate");
        actionMap.put("rotate",new AbstractAction() {
            public void actionPerformed(ActionEvent event) {
                String angle = JOptionPane.showInputDialog(ImageMonkey2.this
                        ,"enter angle between -1 and 1 to rotate"
                        ,"90");
                if(angle!=null){
                    final String title=getTitle();
                    try {
//   todo                     float angleDegrees = Float.parseFloat(angle);
//                        float angleRads = (float)Math.toRadians(angleDegrees);
//                        setTitle("rotating "+angleDegrees+" degrees...");
//
//                        TaskTimer tt = new TaskTimer();
//                        File file = imageFiles.get(index);
//                        String fileName=file.getName();
//                        log.log(Level.INFO,"Creating operation to load image from '" + fileName+"'");
//                        RenderedOp img =  JAI.create("fileload", file.getCanonicalPath());
//                        // h & prop need to be height and width minus chrome
//                        PlanarImage pi = io.rotate(img,angleRads);
//                        setImage(pi);
//                        log.log(Level.INFO,tt.stopString()+" to display "+fileName);
//                        setTitle(getDefaultTitle()+" rotated "+angleDegrees+" degrees");
                    } catch (NumberFormatException e) {
                        setTitle(title);
                        log.log(Level.INFO,"cannot parse "+angle+" into float.. ignoring");
                    } /*catch (IOException e) {
                    e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                    }*/
                }
            }
        });

    }


    private void setupQuickfilesKeys(final InputMap inputMap, final ActionMap actionMap, final List<File> imageFiles) {
        for(final QuickFiles qf: QuickFiles.values()){
            //add actions
            String actionname = "addto_"+qf.getFile().getName();
            KeyStroke keyStroke = KeyStroke.getKeyStroke(qf.getKey(),0);
            inputMap.put(keyStroke,actionname);
            actionMap.put(actionname,new AbstractAction() {
                public void actionPerformed(ActionEvent event) {
                    try {
                        File currF = imageFiles.get(index);
                        final String filename = currF.getName();
                        qf.addName(filename);
                        setTitle(getDefaultTitle()+" appended to "+qf.getFile().getName()+" "+df.format(new Date()));
                    } catch (IOException e) {
                        e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                        JOptionPane.showMessageDialog(ImageMonkey2.this,"cannot write name to file: \n"+e.getMessage(),"add name to file",JOptionPane.ERROR_MESSAGE);
                    }
                }
            });

            //remove actions
            actionname = "removefrom_"+qf.getFile().getName();
            keyStroke = KeyStroke.getKeyStroke(qf.getKey(),KeyEvent.SHIFT_MASK);
            inputMap.put(keyStroke,actionname);
            actionMap.put(actionname,new AbstractAction() {
                public void actionPerformed(ActionEvent event) {
                    if(!qf.getFile().exists()){
                        setTitle(getDefaultTitle()+" no such file:  "+qf.getFile().getName()+" "+df.format(new Date()));
                        return;
                    }
                    try {
                        File currF = imageFiles.get(index);
                        final String filename = currF.getName();
                        int linesRemoved = qf.removeName(filename);
                        setTitle(getDefaultTitle()+" removed from "
                                +qf.getFile().getName()+" (-"+linesRemoved+" lines) "
                                +df.format(new Date()));
                    } catch (IOException e) {
                        e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                        JOptionPane.showMessageDialog(ImageMonkey2.this,"cannot remove name from file: \n"+e.getMessage(),"remove name from file",JOptionPane.ERROR_MESSAGE);
                    }
                }
            });

            //move actions
            actionname = "moveto_"+qf.getFile().getName();
            keyStroke = KeyStroke.getKeyStroke(qf.getKey(),KeyEvent.CTRL_MASK);
            inputMap.put(keyStroke,actionname);
            actionMap.put(actionname,new AbstractAction() {
                public void actionPerformed(ActionEvent event) {
                    try {
                        final String filename = imageFiles.get(index).getName();
                        int linesRemoved =0,filesEdited=0;
                        for(QuickFiles tf1:QuickFiles.values()){
                            int lr = tf1.removeName(filename);
                            if(lr>0){
                                linesRemoved+=lr; filesEdited++;
                            }
                        }
                        qf.addName(filename);
                        setTitle(getDefaultTitle()+" moved to " +
                                qf.getFile().getName()
                                +" (-"+linesRemoved+" lines from "+filesEdited+" files) "
                                +df.format(new Date()));
                    } catch (IOException e) {
                        e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                        JOptionPane.showMessageDialog(ImageMonkey2.this,"cannot move name to file: \n"+e.getMessage(),"remove name from file",JOptionPane.ERROR_MESSAGE);
                    }
                }
            });
        }
    }

    private final DateFormat df = new SimpleDateFormat("HH:mm ss");

    private void showImage(int i) {
        String title=getTitle();
        try {
            final File f = imageFiles.get(i);
            log.fine("show image="+f);
            setTitle("loading "+f.getName()+"...");
            displayImage(f);
            this.index=i;
            setTitle(getDefaultTitle());
            if(i+1<imageFiles.size()){//try to pre-cache the next image
                final File f2 = imageFiles.get(i+1);
                log.fine("cache file="+f2+"?");
                if(!cache.containsKey(f2)){
                    Runnable runnable = new Runnable() {
                        public void run() {
                            log.fine("caching f2="+f2);
                            final Image image = fetchScaledImage(f2,getWidth(),getHeight());
                            image.getWidth(nullObserver);
                            image.getHeight(nullObserver);
                        }
                    };
                    Thread t = new Thread(runnable);
                t.setPriority(Thread.MIN_PRIORITY);
                t.start();
                }

            }

        } catch (IOException e) {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
            setTitle(title);
            JOptionPane.showMessageDialog(this,"cannot show image: \n"+e.getMessage(),"show image",JOptionPane.ERROR_MESSAGE);
        }
    }

    private String getDefaultTitle() {
        return imageFiles.get(index).getName()+" ("+(index+1)+" /"+imageFiles.size()+")";
    }

    private int index=0;
    private void displayImage(File file) throws IOException {
        gotImage=false;
        repaint();
        image = fetchScaledImage(file,getWidth(),getHeight());
        if (image!=null) {
            gotImage=true;
            repaint();
        }
    }
    private boolean gotImage=false;

    private static abstract class IntFetcher implements ImageObserver{
        private int prop=-1;
        public synchronized boolean imageUpdate(Image img, int infoflags, int x, int y, int width, int height) {
            if(gotProperty()) return true;
            prop=getPropertyFromUpdate(img,infoflags,x,y,width,height);
            boolean ret = gotProperty();
                log.fine("notifying all ret="+ret);
                notifyAll();
            return ret;
        }
        private boolean gotProperty(){
            return prop!=-1;
        }
        public int getPropertySynchronous(Image i){
            prop = getPropertyFromImage(i);
            while (!gotProperty()) {
                synchronized(this){
                    try {
                        log.fine("waiting on property update");
                        wait();
                    } catch (InterruptedException e) { }
                }
            }
            log.fine("returning prop="+prop);
            return prop;
        }
        abstract int getPropertyFromImage(Image i);
        abstract int getPropertyFromUpdate(Image img, int infoflags, int x, int y, int width, int height);
    }
    private IntFetcher widthFetcher = new IntFetcher() {
        int getPropertyFromImage(Image i) {
            return i.getWidth(this);
        }

        int getPropertyFromUpdate(Image img, int infoflags, int x, int y, int width, int height) {
            log.fine("prop update width="+width);
            return width;
        }
    };
    private IntFetcher heightFetcher = new IntFetcher() {
        int getPropertyFromImage(Image i) {
            return i.getHeight(this);
        }

        int getPropertyFromUpdate(Image img, int infoflags, int x, int y, int width, int height) {
            log.fine("prop update height="+height);
                        return height;
        }
    };
    private Map<File,Image> cache = new WeakHashMap<File,Image>();
    ImageOperations io = new ImageOperations();
    private Image fetchScaledImage(File f, int maxW, int maxH){
        if(maxW==0 || maxH==0) log.warning("maxW="+maxW+" maxH="+maxH);
        try {
            if(cache.containsKey(f)){
                log.info("found image in cache!");
                return cache.get(f);
            }
            TaskTimer tt = new TaskTimer();
            Image image=Toolkit.getDefaultToolkit().createImage(f.toURL());
            //do any caching and stuff here

            final int imageW=widthFetcher.getPropertySynchronous(image);
            final int imageH=heightFetcher.getPropertySynchronous(image);
            if(imageW==0 || imageH==0){
                log.warning("imagew or h was null! ignoring");
                return null;
            }
            if (log.isLoggable(Level.FINE)) {
                log.fine("paint maxW="+maxW+" maxH="+maxH+ " imageW="+imageW+" imageH="+imageH);
            }
            //we want the scaled image to fit within our box of drawWxdrawH
            final int constVal = scaleType.isSet()?scaleType.getArgValue().getConstVal():Image.SCALE_FAST;
            image = io.scale(image,imageW,imageH,maxW,maxH,true,constVal);
            log.info(tt.stopString()+" to load image and getScaledInstance");
            cache.put(f,image);
            return image;
        } catch (MalformedURLException e) {
            log.log(Level.WARNING, "exception", e);
            return null;
        }
    }

    private Image image;


}

