package uk.co.hcsoftware.images.gui;

import com.sun.media.jai.widget.DisplayJAI;
import ns.TaskTimer;
import uk.co.hcsoftware.images.tools.ImageOperations;

import javax.media.jai.JAI;
import javax.media.jai.PlanarImage;
import javax.media.jai.RenderedOp;
import javax.swing.*;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.io.*;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.logging.Logger;
import java.util.logging.Level;

/**
 * image viewer app for flicking through images/thumbnails quickly.
 *
 * Tried windows native libs and didn't seem any faster to load/scale/display..
 * was taking ~937ms to display ~1.5mb png files
 * java version was taking ~1s
 *
 *
 */
public class ImageViewer extends JFrame{
    private static final Logger log = Logger.getLogger(ImageViewer.class.getName());
    private static class UsageException extends Exception{
        public UsageException(String message) {
            super(message);
        }
    };
    public static void main(String args[]) {
        try {
            if(args.length==0){
                throw new UsageException("invalid args");
            }
            List<File> imageFiles = new ArrayList<File>(args.length);
            for (int i = 0; i < args.length; i++) {
                String arg = args[i];
                imageFiles.add(new File(arg));
            }
            new ImageViewer(imageFiles);
            return;
        } catch(UsageException e){
            ;
        }catch (Exception e) {
            e.printStackTrace();
        }
        log.log(Level.INFO,getUsage());
    }

    private static String getUsage() {
        return ImageViewer.class.getName()+" imagefiles" +
                "\n\tdisplays a window with the first image." +
                "\n\timages are scaled to fit the window" +

                "\nNavigating:"+
                "\n\tPgDn=next image" +
                "\n\tPgUp=prev image" +
                "\n\tHOME=first image" +
                "\n\tEND=last image" +
                "\n\tr=reload current image" +

                "\nQuickFiles:" +
                "(textfiles holding 1 filename per line, created in workingdir, called imagenames_${keyname})"+
                "\n\tF1-F5=add curr imagename to quickfile" +
                "\n\tSHIFT+[F1-F5]= remove curr imagename from quickfile" +
                "\n\tCTRL+[F1-F5]= move curr imagename to quickfile (remove from all other quickfiles)" +

                "\nEditing" +
                "\n\tCTRL+r= reload from file and rotate about centre" +
                "\n\nOptimisation:" +
                "\nBased on JAI 1.2, see http://java.sun.com/products/java-media/jai for details on how" +
                "to optimize for your system";
    }

    final List<File> imageFiles;
    enum QuickFiles{
        F1("imagenames_f1", KeyEvent.VK_F1),F2("imagenames_f2", KeyEvent.VK_F2),F3("imagenames_f3", KeyEvent.VK_F3),F4("imagenames_f4", KeyEvent.VK_F4),F5("imagenames_f5", KeyEvent.VK_F5);
        private final int key;

        public File getFile() {
            return f;
        }

        public int getKey() {
            return key;
        }

        private final File f;
        QuickFiles(String s, int key){
            this.key = key;
            this.f = new File(s);
        }
        private void addName(String name) throws IOException {
            FileWriter fw = new FileWriter(f,true);
            fw.write(name+"\n");
            fw.close();
        }
        private int removeName(String filename) throws IOException {
            if(!f.exists()){
                return 0;
            }
            LineNumberReader lnr = new LineNumberReader(new FileReader(f));
            List<String> filenames = new ArrayList<String>(50);
            String line;
            int removed=0;
            while((line=lnr.readLine())!=null){
                if(!line.equals(filename)) filenames.add(line);
                else removed++;
            }
            lnr.close();
            FileWriter fw = new FileWriter(f,false);
            for(String fname: filenames){
                fw.write(fname+"\n");
            }
            fw.close();
            return removed ;
        }

    }
    //todo: cropping: should crop the _original file_ and reload (not the scaled version)
    public ImageViewer(final List<File> imageFiles) {
        super("imageviewer");
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        this.imageFiles=imageFiles;
        // getContentPane().setLayout(new BorderLayout(2,0));
        add(new JScrollPane(displayPanel));

        setupKeys(imageFiles);

        setSize(600,400);
        setVisible(true);
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                showImage(0);
            }
        });
    }

    private void setupKeys(final List<File> imageFiles) {
        final InputMap inputMap = displayPanel.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW);
        final ActionMap actionMap = displayPanel.getActionMap();

        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_DOWN,0),"next");
        actionMap.put("next",new AbstractAction() {
            public void actionPerformed(ActionEvent event) {
                showImage(index+1>=imageFiles.size()?0:index+1);
            }
        });
        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_UP,0),"previous");
        actionMap.put("previous",new AbstractAction() {
            public void actionPerformed(ActionEvent event) {
                showImage(index==0?imageFiles.size()-1:index-1);
            }
        });
        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_HOME,0),"first");
        actionMap.put("first",new AbstractAction() {
            public void actionPerformed(ActionEvent event) {
                showImage(0);
            }
        });
        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_END,0),"last");
        actionMap.put("last",new AbstractAction() {
            public void actionPerformed(ActionEvent event) {
                showImage(imageFiles.size()-1);
            }
        });
        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_R,0),"reload");
        actionMap.put("reload",new AbstractAction() {
            public void actionPerformed(ActionEvent event) {
                showImage(index);
            }
        });

        setupQuickfilesKeys(inputMap, actionMap, imageFiles);

        //image edit actions
        inputMap.put(KeyStroke.getKeyStroke(KeyEvent.VK_R,KeyEvent.CTRL_MASK),"rotate");
        actionMap.put("rotate",new AbstractAction() {
            public void actionPerformed(ActionEvent event) {
                String angle = JOptionPane.showInputDialog(ImageViewer.this
                        ,"enter angle between -1 and 1 to rotate"
                ,"90");
                if(angle!=null){
                    final String title=getTitle();
                    try {
                        float angleDegrees = Float.parseFloat(angle);
                        float angleRads = (float)Math.toRadians(angleDegrees);
                        setTitle("rotating "+angleDegrees+" degrees...");

                        TaskTimer tt = new TaskTimer();
                        File file = imageFiles.get(index);
                        String fileName=file.getName();
                        log.log(Level.INFO,"Creating operation to load image from '" + fileName+"'");
                        RenderedOp img =  JAI.create("fileload", file.getCanonicalPath());
                        // h & w need to be height and width minus chrome
                        PlanarImage pi = io.rotate(img,angleRads);
                        setImage(pi);
                        log.log(Level.INFO,tt.stopString()+" to display "+fileName);
                        setTitle(getDefaultTitle()+" rotated "+angleDegrees+" degrees");
                    } catch (NumberFormatException e) {
                        setTitle(title);
                        log.log(Level.INFO,"cannot parse "+angle+" into float.. ignoring");
                    } catch (IOException e) {
                        e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                    }
                }
            }
        });

    }



    private void setupQuickfilesKeys(final InputMap inputMap, final ActionMap actionMap, final List<File> imageFiles) {
        for(final QuickFiles qf: QuickFiles.values()){
            //add actions
            String actionname = "addto_"+qf.getFile().getName();
            KeyStroke keyStroke = KeyStroke.getKeyStroke(qf.getKey(),0);
            inputMap.put(keyStroke,actionname);
            actionMap.put(actionname,new AbstractAction() {
                public void actionPerformed(ActionEvent event) {
                    try {
                        File currF = imageFiles.get(index);
                        final String filename = currF.getName();
                        qf.addName(filename);
                        setTitle(getDefaultTitle()+" appended to "+qf.getFile().getName()+" "+df.format(new Date()));
                    } catch (IOException e) {
                        e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                        JOptionPane.showMessageDialog(ImageViewer.this,"cannot write name to file: \n"+e.getMessage(),"add name to file",JOptionPane.ERROR_MESSAGE);
                    }
                }
            });

            //remove actions
            actionname = "removefrom_"+qf.getFile().getName();
            keyStroke = KeyStroke.getKeyStroke(qf.getKey(),KeyEvent.SHIFT_MASK);
            inputMap.put(keyStroke,actionname);
            actionMap.put(actionname,new AbstractAction() {
                public void actionPerformed(ActionEvent event) {
                    if(!qf.getFile().exists()){
                        setTitle(getDefaultTitle()+" no such file:  "+qf.getFile().getName()+" "+df.format(new Date()));
                        return;
                    }
                    try {
                        File currF = imageFiles.get(index);
                        final String filename = currF.getName();
                        int linesRemoved = qf.removeName(filename);
                        setTitle(getDefaultTitle()+" removed from "
                                +qf.getFile().getName()+" (-"+linesRemoved+" lines) "
                                +df.format(new Date()));
                    } catch (IOException e) {
                        e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                        JOptionPane.showMessageDialog(ImageViewer.this,"cannot remove name from file: \n"+e.getMessage(),"remove name from file",JOptionPane.ERROR_MESSAGE);
                    }
                }
            });

            //move actions
            actionname = "moveto_"+qf.getFile().getName();
            keyStroke = KeyStroke.getKeyStroke(qf.getKey(),KeyEvent.CTRL_MASK);
            inputMap.put(keyStroke,actionname);
            actionMap.put(actionname,new AbstractAction() {
                public void actionPerformed(ActionEvent event) {
                    try {
                        final String filename = imageFiles.get(index).getName();
                        int linesRemoved =0,filesEdited=0;
                        for(QuickFiles tf1:QuickFiles.values()){
                            int lr = tf1.removeName(filename);
                            if(lr>0){
                                linesRemoved+=lr; filesEdited++;
                            }
                        }
                        qf.addName(filename);
                        setTitle(getDefaultTitle()+" moved to " +
                                qf.getFile().getName()
                                +" (-"+linesRemoved+" lines from "+filesEdited+" files) "
                                +df.format(new Date()));
                    } catch (IOException e) {
                        e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                        JOptionPane.showMessageDialog(ImageViewer.this,"cannot move name to file: \n"+e.getMessage(),"remove name from file",JOptionPane.ERROR_MESSAGE);
                    }
                }
            });
        }
    }

    private final DateFormat df = new SimpleDateFormat("HH:mm ss");

    private void showImage(int i) {
        String title=getTitle();
        try {
            final File f = imageFiles.get(i);
            setTitle("loading "+f.getName()+"...");
            setImage(f);
            this.index=i;
            setTitle(getDefaultTitle());
        } catch (IOException e) {
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
            setTitle(title);
            JOptionPane.showMessageDialog(this,"cannot show image: \n"+e.getMessage(),"show image",JOptionPane.ERROR_MESSAGE);
        }
    }

    private String getDefaultTitle() {
        return imageFiles.get(index).getName()+" ("+(index+1)+" /"+imageFiles.size()+")";
    }

    private DisplayJAI displayPanel = new DisplayJAI();
    ImageOperations io = new ImageOperations();
    private int index=0;
    private void setImage(File file) throws IOException {
        TaskTimer tt = new TaskTimer();
        String fileName=file.getName();
        log.log(Level.INFO,"Creating operation to load image from '" + fileName+"'");
        RenderedOp img =  JAI.create("fileload", file.getCanonicalPath());
        setImage(img);
        log.log(Level.INFO,tt.stopString()+" to display "+fileName);
    }
    private void setImage(PlanarImage pi) {
        // h & w need to be height and width minus chrome
        pi=io.scale(pi, getContentPane().getWidth()-10, getContentPane().getHeight()-10, true);
        displayPanel.set(pi);
    }
}

