package uk.co.hcsoftware.images.gui.map;

import com.sun.media.jai.widget.DisplayJAI;
import org.apache.commons.cli.*;
import uk.co.hcsoftware.images.tools.ImageOperations;

import javax.media.jai.JAI;
import javax.media.jai.PlanarImage;
import javax.print.*;
import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.standard.OrientationRequested;
import javax.print.event.PrintJobAdapter;
import javax.print.event.PrintJobEvent;
import javax.swing.*;
import java.awt.*;
import java.awt.geom.AffineTransform;
import java.awt.image.renderable.ParameterBlock;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.io.*;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;
import java.util.logging.Level;

public class MainClass {
    private static final Logger log = Logger.getLogger(MainClass.class.getName());

    //map tiles
    public static final Option mapsroot = new Option("r","mapsroot",true,"root directory of mapfiles named like: ${root}/${gridletter}/${6figuregrid}.png");
    public static final Option findmaps = new Option("f","findmaps",true,"look for mapfiles under this dir. Files must be named e.g. TQ100900.png");

    //ref resolving
    public static final Option postcodefile = new Option("p","postcodefile",true,"file with postcode-(x,y) mappings");
    public static final Option namefile = new Option("n","namefile",true,"file with name-(x,y) mappings");

    //image attribs
    public static final Option border = new Option("m","margin",true,"extra metres around bounding box to include in map (default=1000)");
    public static final Option chop = new Option("c","chop",true,"x,y chop the output map up into submaps each x by y pixels in size. If saving, save to multiple files, if displaying display multiple files");
    public static final Option screen = new Option("screen",false,"scale to maxpect size that will fit on screen, can be combined with scale, causes chop to be ignored");
    public static final Option scale = new Option("x",true,"(n) scale map down to 1/n original size");
    public static final Option maxw = new Option("mw",true,"max width of map in pixels (scale, keeping aspect ratio)");
    public static final Option maxh = new Option("mh",true,"max height of map in pixels (scale, keeping aspect ratio)");

    public static final Option verbose = new Option("v","verbose",false,"does nothing -here for legacy reasons use -Djava.util.logging.config.file=.. for logging");

    //output
    public static final Option outformat = new Option("outformat",true,"format for output images e.g. jpeg (default=png) see javax.media.jai.operator.FileStoreDescriptor for supported formats");//todo:
    public static final Option save = new Option("s",true,"(filename) save the map to a file");
    public static final Option stdout = new Option("o",false,"write map to std.out");
    public static final Option display = new Option("d","display",false,"display the map onscreen");
    public static final Option print = new Option("print",true,"(p | l) for portrait or landscape: Print to the default printservice, chops into pages, if chop is specified it is ignored. WARNING: a large map can use up lots of pages... good idea to specify verbose option (a4 lanscape accommodate ~ 451px x 697px)");

    //decorations
    public static final Option rings = new Option("rings",false,"print rings around each point");
    public static final Option lines = new Option("lines",false,"print lines connecting point1-pointn and each point in between");
    public static final Option distances = new Option("distances",false,"print distances point1-pointn and each point in between @ the midpoint; implies -lines");
    public static final Option times = new Option("times",false,"print times in which a good club cyclist should be able to cover the distance point1-pointn; implies -lines");


    private static final Options availableOptions = new Options();
    private static CommandLine cmd;


    static{
        try {
            Field[] fields = MainClass.class.getDeclaredFields();
            for (int i = 0; i < fields.length; i++) {
                Field field = fields[i];
                if(field.getType()==Option.class){ //all fields that are Option s
                    MainClass.availableOptions.addOption((Option)field.get(null));
                }
            }
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        }
    }



    private static void usage(){
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp(80
                , MapFactory.class.getName()+" [opts] point1 [point2.. n]"
                ,"splices map tiles to create a new map the area of which is the smallest rectangle" +
                " that containts all the points\n" +
                "points can be any one of: " +
                "\n  UK postcode (e.g. E14)" +
                "\n  Landranger grid (e.g. TQ315762)" +
                "\n  x,y metres from SW of UK (e.g. 530000,170000) (see -p)" +
                "\n  name in mapref file (see -n option)"
                , availableOptions
                ,"Example:\n  java -jar "+MapFactory.class.getName().substring(MapFactory.class.getName().lastIndexOf('.')+1)
                +".jar -v E14 W5\n" +
                "generates a map with E14, W5 and all the points between"
                ,false
        );
    }

    public static void main(String[] args) {
        try {
            cmd = new GnuParser().parse(availableOptions,args);
            if(log.isLoggable(Level.FINE)){
                Option[] opts = cmd.getOptions();
                StringBuffer sb = new StringBuffer();
                for (int i = 0; i < opts.length; i++) {
                    Option opt = opts[i];
                    sb.append(opt.getOpt()+"="+opt.getValue());
                    if(i+1<opts.length) sb.append(" ");
                }
                log.log(Level.FINE,"options: "+sb.toString());
            }
            final String[] arguments = cmd.getArgs();
            if(arguments.length>0){
                new MainClass().run(arguments);
                return;
            }
        }  catch(MissingOptionException e){
            System.err.println("Missing option(s): "+e.getMessage());
        } catch (ParseException e) {
            e.printStackTrace();
        } catch (MapFactory.NoMapFilesException e) {
            log.log(Level.WARNING,"I couldn't find any map files: "+e.getMessage(),e);
        } catch (InvalidArgException e) {
            System.err.println("invalid argument value="+e.getArgValue()+" for argument="+e.getArgName()+": "+e.getMessage());
        }
        usage();
        //  System.exit(1);
    }

    private void run(String[] arguments) throws MapFactory.NoMapFilesException, InvalidArgException {
        final RefResolver rr = new RefResolver(
                MainClass.cmd.getOptionValue(MainClass.postcodefile.getOpt())
                ,MainClass.cmd.getOptionValue(MainClass.namefile.getOpt()));
        MapFactory mb = new MapFactory();

        readOptions(mb);
        try {
            java.util.List<Point> l = new ArrayList<Point>(arguments.length);
            for (int i1 = 0; i1 < arguments.length; i1++) {
                String argument = arguments[i1];
                l.add(rr.getPoint(argument));
            }
            java.util.List<Point> points = l;

            if(points.size()==0){
                System.err.println("no valid points supplied");
                usage();
                return;
            }
            log.log(Level.FINE,"points = " + points);

            List<PlanarImage> pis = mb.createMap(l);

            if (cmd.hasOption(save.getOpt())) {
                saveFiles(cmd.getOptionValue(save.getOpt()), pis);
            }
            if (cmd.hasOption(display.getOpt())) {
                displayFrames(pis);
            }
            if (cmd.hasOption(print.getOpt())) {
                String v = cmd.getOptionValue(print.getOpt());
                final int i;
                if("p".equals(v)){
                    i=PageFormat.PORTRAIT;
                }
                else if("l".equals(v)){
                    i=PageFormat.LANDSCAPE;
                }
                else throw new InvalidArgException(print.getOpt(),v,"must be p or l (for portrait or landscape)");
                print(pis,i);
            }
            if (cmd.hasOption(stdout.getOpt())) {
                for (PlanarImage pi:pis) {
                    log.log(Level.INFO,"writing to stdout");
                    ParameterBlock pb = new ParameterBlock();
                    pb.addSource(pi);
                    pb.add(System.out);
                    final String of = cmd.getOptionValue(outformat.getOpt(),"png");
                    pb.add(of);
                    JAI.create("encode", pb, null);
                }
            }
            //  System.exit(0);
        }  catch (MapFactory.BadReferenceException e) {
            log.log(Level.WARNING,"I couldn't understand reference="+e.getReference(),e);
        } catch (CannotResolveReferenceException e) {
            log.log(Level.WARNING,"I couldn't resolve reference="+e.getReference(),e);
        } catch (PrintException e) {
            log.log(Level.WARNING,"printing error",e);
        }
    }

    private static String getMapsRootFromMaprefFile() {
        String nfile=cmd.getOptionValue(namefile.getOpt());
        if(nfile!=null){
            File f = new File(nfile);
            LineNumberReader lnr = null;
            try {
                lnr = new LineNumberReader(new FileReader(f));
            } catch (FileNotFoundException e) {
                e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
            }
            try {
                String line;
                while((line=lnr.readLine())!=null){
                    final int i = line.indexOf("root:");
                    if(i!=-1){
                        return line.substring(i+1);
                    }
                }
            } catch(Exception e){ e.printStackTrace();}
            finally{
                try {
                    lnr.close();
                } catch (IOException e) {
                    e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
                }
            }
        }
        return null;
    }

    private void print(List<PlanarImage> pis, final int orientation) throws PrintException {
        final PlanarImage pi = pis.get(0);
        PrintService service = PrintServiceLookup.lookupDefaultPrintService();
        log.log(Level.FINE,"creating print job format="+(orientation==PageFormat.PORTRAIT?"portrait":"landscape")+"...");
        DocPrintJob job = service.createPrintJob();
        Printable pr = new Printable() {
            final ImageOperations io = new ImageOperations();
            private List<PlanarImage> images;
            public int print(Graphics graphics, PageFormat pf, int pageIndex) {
               // pf.setOrientation(orientation);
                if(images==null){
                    final int wPixels = (int)pf.getImageableWidth();
                    final int hPixels = (int) pf.getImageableHeight();
                        log.log(Level.INFO,"printable size= " + hPixels+"px x " + wPixels+"px");
                    images= io.chop(pi,wPixels,hPixels, Color.white);
                    log.log(Level.INFO,"image chopped into "+images.size()+" pages");
                }
                if(pageIndex>=images.size()) return Printable.NO_SUCH_PAGE;
                PlanarImage image= images.get(pageIndex);

                Graphics2D g2 = (Graphics2D) graphics;
                final AffineTransform at = new AffineTransform();
                at.setToTranslation(pf.getImageableX(), pf.getImageableY());
                g2.drawRenderedImage(image,at);
                return Printable.PAGE_EXISTS;
            }
        };
        Doc doc = new SimpleDoc(pr, DocFlavor.SERVICE_FORMATTED.PRINTABLE, null);
        job.addPrintJobListener(new PrintJobAdapter() {
            public void printJobCompleted(PrintJobEvent pje) {log.log(Level.INFO,"print job completed");}
            public void printJobFailed(PrintJobEvent pje) {log.log(Level.INFO,"print job failed");}
            public void printJobCanceled(PrintJobEvent pje) {log.log(Level.INFO,"print job cancelled");}
        });
        PrintRequestAttributeSet aset = new HashPrintRequestAttributeSet();
        if (orientation==PageFormat.PORTRAIT) {
            aset.add(OrientationRequested.PORTRAIT);
        } else {
            aset.add(OrientationRequested.LANDSCAPE);
        }
        job.print(doc, aset);
    }

    private void displayFrames(List<PlanarImage> pis) {
        Point loc=null;
        int j=0;
        for (PlanarImage pi:pis) {
            DisplayJAI displayPanel = new DisplayJAI();
            displayPanel.set(pi);
            JFrame jf = new JFrame(MapFactory.class.getName()+" "+j);
            jf.add(new JScrollPane(displayPanel));
            if (pis.size()==1) {
                jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
            }
            if(MainClass.cmd.hasOption(MainClass.screen.getOpt())){
                jf.pack();
            }
            else jf.setSize(800,600);
            if(loc!=null){
                loc.translate(10,10);
                jf.setLocation(loc);
            }
            jf.setVisible(true);
            loc=jf.getLocation();
            j++;
        }
    }

    private void saveFiles(final String sval, List<PlanarImage> pis) {
        if(sval!=null){
            String savefile=sval;

            String ext="";
            String allButExt=savefile;
            int i=savefile.indexOf('.');//= savefile.substring()
            if(i!=-1){
                ext=savefile.substring(i);
                allButExt=savefile.substring(0,i);
            }
            int j=0;
            for (PlanarImage pi:pis) {
                if(pis.size()>0){
                    savefile=allButExt+j+ext;
                }
                File f = new File(savefile);
                    log.log(Level.INFO,"saving to "+savefile);

                ParameterBlock pb = new ParameterBlock();
                pb.addSource(pi);
                pb.add(f.getAbsolutePath());
                final String of = cmd.getOptionValue(outformat.getOpt(),"png");
                pb.add(of);
                JAI.create("filestore", pb, null);
                j++;
            }

        }
    }

    private void readOptions( MapFactory mb) throws InvalidArgException, MapFactory.NoMapFilesException {

        //set options
        mb.setDrawRings(MainClass.cmd.hasOption(MainClass.rings.getOpt()));
        mb.setDrawLines(MainClass.cmd.hasOption(MainClass.lines.getOpt()));
        mb.setDrawDistances(MainClass.cmd.hasOption(MainClass.distances.getOpt()));
        mb.setDrawTimes(MainClass.cmd.hasOption(MainClass.times.getOpt()));
       // mb.setVerbose(MainClass.cmd.hasOption(MainClass.verbose.getOpt()));

        if(MainClass.cmd.hasOption(MainClass.screen.getOpt()) ){
            mb.setScaleToScreen(true);
        }
        if(MainClass.cmd.hasOption(MainClass.scale.getOpt()) ){
            String s=MainClass.cmd.getOptionValue(MainClass.scale.getOpt());
            try {
                mb.setScaleValue(Integer.parseInt(s));
            } catch (NumberFormatException e) {
                throw new InvalidArgException(MainClass.scale.getOpt(),s,"cannot convert to number");
            }
        }

        if(MainClass.cmd.hasOption(MainClass.maxw.getOpt()) || MainClass.cmd.hasOption(MainClass.maxh.getOpt())){
            String w=MainClass.cmd.getOptionValue(MainClass.maxw.getOpt());
            String h=MainClass.cmd.getOptionValue(MainClass.maxh.getOpt());
            if(w!=null){
                try {
                    mb.setMaxW(Integer.parseInt(w));
                } catch (NumberFormatException e) {
                    throw new InvalidArgException(MainClass.maxw.getOpt(),w,"cannot convert to number");
                }
            }
            if(h!=null){
                try {
                    mb.setMaxH(Integer.parseInt(h));
                } catch (NumberFormatException e) {
                    throw new InvalidArgException(MainClass.maxh.getOpt(),h,"cannot convert to number");
                }
            }
        }
        if(MainClass.cmd.hasOption(MainClass.chop.getOpt()) && !MainClass.cmd.hasOption(MainClass.print.getOpt())){
            String a=MainClass.cmd.getOptionValue(MainClass.chop.getOpt());
            try {
                String[] xy=a.split(",");
                int w=Integer.parseInt(xy[0]);
                int h=Integer.parseInt(xy[1]);
                mb.setChop(w,h);
            } catch (NumberFormatException e) {
                throw new InvalidArgException(MainClass.chop.getOpt(),a,"cannot convert to number");
            } catch (ArrayIndexOutOfBoundsException e) {
                throw new InvalidArgException(MainClass.chop.getOpt(),a,"should be x,y format");
            }
        }

        String mapsRoot=cmd.getOptionValue(mapsroot.getOpt());
        //if mapsroot is null, have a look in the namefile just in case..
        if(mapsRoot==null){
            mapsRoot=getMapsRootFromMaprefFile();
        }

        if(cmd.hasOption(findmaps.getOpt())){
            //mb.getImaginaryTiles(cmd.getOptionValue(findmaps.getOpt()));
            String findMaps = cmd.getOptionValue(findmaps.getOpt());
            File f = new File(findMaps);
            if(!f.exists()) throw new MapFactory.NoMapFilesException("root file="+f+" doesn't exist");
            if(!f.isDirectory()) throw new MapFactory.NoMapFilesException("root file="+f+" isn't a directory");
            mb.findTiles(f);
            log.log(Level.INFO,"found "+mb.mapTiles.size()+" map tiles under "+findMaps);
        }
        else if(mapsRoot==null) throw new MapFactory.NoMapFilesException("root file/find dir not supplied");

        if(mapsRoot!=null){
            mb.setMapsRoot(mapsRoot);//getImaginaryTiles(mapsRoot);
        }
        if(cmd.hasOption(border.getOpt())){
            String bval = cmd.getOptionValue(border.getOpt());
            try {
                mb.setMargin(Integer.parseInt(bval));
            } catch (NumberFormatException e) {
                throw new InvalidArgException(border.getOpt(),bval,"can't convert to number");
            }
        }
    }

}
