package uk.co.hcsoftware.images.gui.map;

import ns.TaskTimer;
import uk.co.hcsoftware.images.tools.ImageOperations;

import javax.media.jai.PlanarImage;
import javax.media.jai.TiledImage;
import java.awt.*;
import java.awt.geom.AffineTransform;
import java.io.File;
import java.math.BigDecimal;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.*;
import java.util.List;
import java.util.logging.Logger;
import java.util.logging.Level;

public class MapFactory {
    private static final Logger log = Logger.getLogger(MapFactory.class.getName());

    private static final Color DECORATIONS_COLOR = new Color(0,0,0,200);
    private static final BasicStroke DECORATIONS_STROKE = new BasicStroke(6);
    private static final Font DECORATIONS_FONT = new Font("SansSerif",Font.BOLD,48);

    //options
    private boolean drawRings;
    private boolean drawLines;
    private int scaleValue=-1;
    private int maxW=-1,maxH=-1;
    private int chopX=-1,chopY=-1;
    private int margin=1000;
    private boolean distances;
    private boolean times;
    private boolean scaleToScreen;


    public void setMaxW(int maxW) {
        this.maxW = maxW;
    }

    public void setMaxH(int maxH) {
        this.maxH = maxH;
    }


    public void setDrawRings(boolean drawRings) {
        this.drawRings = drawRings;
    }

    public void setDrawLines(boolean drawLines) {
        this.drawLines = drawLines;
    }

    public void setScaleValue(int scaleValue) {
        this.scaleValue = scaleValue;
    }

    private void findFiles(File root, List<File> addToList){
        for(File f:root.listFiles()){
            if(f.isDirectory()){
                findFiles(f,addToList);
            }
            else addToList.add(f);
        }
    }
    private void getImaginaryTiles(String rootFile) throws NoMapFilesException {
        final File rootF = new File(rootFile);
        for(OsGrid.Letter let:grid.getLetters()){
            //each square has 100 subsquares named from 000000 to 900900
            final File letterDir=new File(rootF,let.getName());
            NumberFormat nf = new DecimalFormat("000");
            for(int i=0;i<10;i++){
                for(int j=0;j<10;j++){
                    int xDecaMetres=i*100;
                    int yDecaMetres=j*100;
                    int xMetres=xDecaMetres*100;
                    int yMetres=yDecaMetres*100;
                    String fname=let.getName()+nf.format(xDecaMetres)+nf.format(yDecaMetres)+".png";
                    File f = new File(letterDir,fname);
                    MapTile mt = new MapTile(let.x+xMetres,let.y+yMetres,10000,10000,f);
                    mapTiles.add(mt);
                }
            }
        }
        if(mapTiles.size()==0) throw new NoMapFilesException("no valid map imagefiles found under root="+rootFile+"" +
                " imagefiles must be an image file the prefix of which (everything before the first dot) is an OS 6-figure gridref denoting the bottom RH corner of the tile (e.g. TQ100900)");
    }
    public void findTiles(File mapsRootFile) throws NoMapFilesException {
        //create the maptiles
        List<File> l = new ArrayList<File>();
        findFiles(mapsRootFile,l);
        // mapTiles=new ArrayList<MapTile>(l.size());
        for(File f1:l){
            String prefix=f1.getName();
            if(prefix.indexOf('.')!=-1){
                prefix=prefix.substring(0,prefix.indexOf('.'));
            }
            //prefix is a 6-figure gridref, so convert this to an x,y (bl relative)
            try {
                String let=prefix.substring(0,2);
                OsGrid.Letter l1 = grid.getLetter(let);
                int precision = (prefix.length()-2)/2;
                String x=prefix.substring(2,precision+2);
                String y=prefix.substring(precision+2);
                int xMetres=Integer.parseInt(x)*100;
                int yMetres=Integer.parseInt(y)*100;
                MapTile mt = new MapTile(l1.x+xMetres,l1.y+yMetres,10000,10000,f1);
                mapTiles.add(mt);
            } catch (Exception e) {
                System.err.println("ignoring file="+f1+" as cannot resolve gridref="+prefix+": "+e.getMessage());
            }
        }
        if(mapTiles.size()==0) throw new NoMapFilesException("no valid map imagefiles found under root="+mapsRootFile.getName()+"" +
                " imagefiles must be an image file the prefix of which (everything before the first dot) is an OS 6-figure gridref denoting the bottom RH corner of the tile (e.g. TQ100900)");
    }

    public void setChop(int w, int h) {
        this.chopX=w; this.chopY=h;
    }

    public void setMapsRoot(String mapsRoot) throws NoMapFilesException {
        getImaginaryTiles(mapsRoot);
    }

    public void setMargin(int border) {
        this.margin = border;
    }

    public void setDrawDistances(boolean b) {
        this.distances=b;
    }

    public void setDrawTimes(boolean b) {
        this.times=b;
    }

    public void setScaleToScreen(boolean b) {
        this.scaleToScreen=b;
    }

    static class NoMapFilesException extends Exception{
        public NoMapFilesException(String message) {
            super(message);
        }
    }

    List<MapTile> mapTiles=new ArrayList<MapTile>();

    public List<PlanarImage> createMap(List<Point> metrePoints) throws NoMapFilesException {
        Rectangle box = new Rectangle(metrePoints.get(0));
        for(Point p:metrePoints.subList(1,metrePoints.size())){
            box.add(p);
        }
        box.grow(margin,margin);

        TaskTimer tt = new TaskTimer();
        java.util.List<PlanarImage> pis = createMap(box,metrePoints);

        log.info(tt.stopString()+" to create map");
        return pis;
    }

    /**
     *
     * @param box boundingBox in metres, relative to bottom left of grid
     * @param circlePoints
     * @return
     * @throws NoMapFilesException
     */
    private List<PlanarImage> createMap(Rectangle box,List<Point> circlePoints) throws NoMapFilesException {
        log.fine("box = " + box);
        //1.
        List<MapTile> imageTiles=new ArrayList<MapTile>();
        //find min and max x and y (in pixels)

        Rectangle pixClip = new Rectangle(
                metresToPixels(box.x)
                ,metresToPixels(box.y)
                ,metresToPixels(box.width)
                ,metresToPixels(box.height)
        );//.getPictangle();
        for(MapTile mt:mapTiles){
            if(box.intersects(mt)){
                imageTiles.add(mt);
            }
        }
        if(imageTiles.size()==0){
            throw new NoMapFilesException("0 tiles overlap the requested points");
        }
        log.fine("pixClip = " + pixClip);
        //thanks to  jkyraz <jkyraz@TIN.IT>
        //http://archives.java.sun.com/cgi-bin/wa?A2=ind0405&L=jai-interest&F=&S=&P=13151
        PlanarImage image=null;
        Graphics2D g=null;
        for(MapTile mt:imageTiles){
            try {
                final PlanarImage pi=mt.getImage();
                if(image==null){
                    TiledImage ti = new TiledImage(0,0,pixClip.width,pixClip.height
                            ,0,0,pi.getSampleModel(),pi.getColorModel());
                    g = ti.createGraphics();
                    image=ti;
                }
                AffineTransform a = new AffineTransform();
                //work out the offset from pixClip
                int xOff_m = mt.x-box.x;

                int downlap = box.height+box.y-mt.y;
                int uplap = mt.height-downlap;
                int yOff = metresToPixels(-uplap);

                int xOff = metresToPixels(xOff_m);

                a.setToTranslation(xOff,yOff);
                g.drawRenderedImage(pi,a);
            } catch (NoFileForTileException e) {
                log.log(Level.WARNING,"couldn't find tile file="+e.getImagefile()+" for tile="+e.getMapTile()+": ignoring..");
            }
        }
        if(image==null || g==null){
            throw new NoMapFilesException("0 tiles found");
        }
        if(drawRings){
            g.setColor(DECORATIONS_COLOR);
            g.setStroke(DECORATIONS_STROKE);
            for(Point p:circlePoints){
                Point pixelPoint=getPointPixels(p,box);
                g.drawOval(pixelPoint.x-20,pixelPoint.y-20,40,40);
            }
        }
        if(drawLines || distances || times){
            g.setColor(DECORATIONS_COLOR);
            g.setStroke(DECORATIONS_STROKE);
            //get each pair of points
            Iterator<Point> it = circlePoints.iterator();
            if(it.hasNext()){
                Point p = getPointPixels(it.next(),box);
                while(it.hasNext()){
                    Point p1=getPointPixels(it.next(),box);
                    g.drawLine(p.x,p.y,p1.x,p1.y);
                    if(distances || times){
                        //find the midpoint
                        Point p2 = findMidPoint(p,p1);

                        g.setFont(DECORATIONS_FONT);
                        StringBuffer sb = new StringBuffer();
                        //find the distnace
                            float metres= pixelsToMetres((int) p.distance(p1));
                        if(distances){
                            sb.append(formatMetres(metres));
                        }
                        if(times){
                            if(sb.length()>0) sb.append(' ');
                            long kmh=32;
                            long mh=kmh*1000;
                            long ms=mh/(60*60);
                            long seconds = (long) (metres/ms);
                            sb.append(formatSeconds(seconds));
                        }

                        g.drawString(sb.toString(),p2.x,p2.y);
                    }
                    p=p1;
                }
            }
        }
        if(scaleValue!=-1 ){
            int mw = image.getWidth()/scaleValue;
            log.info("reducing to 1/"+scaleValue+" original size");
            image= io.scale(image,mw,-1,true);
        }

        if(maxW!=-1 || maxH!=-1 ){
            log.info("scaling to maxW="+maxW+" maxH="+maxH);
            image= io.scale(image,maxW,maxH,true);
        }

        if(scaleToScreen){
            Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
            log.info("scaling to primary screensize="+d);
            image= io.scale(image,d.width,d.height,true);
        }
        else if(chopX!=-1){
            log.info("chopping to "+chopX+"x"+chopY+" pixel submaps..");
            final List<PlanarImage> planarImages = io.chop(image,chopX,chopY, Color.white);
            log.info(".."+planarImages.size()+" submaps");
            return planarImages;
        }
        return Arrays.asList(new PlanarImage[]{image});
    }

    private String formatSeconds(long secs) {
        long seconds = secs;
        final int secondsPerDay = 24*60*60;
        long days=seconds/secondsPerDay;
        seconds-=(days*secondsPerDay);
        final int secondsPerHour = 60*60;
        long hours = seconds/secondsPerHour;
        seconds-=(hours*secondsPerHour);
        final int secondsPerMinute = 60;
        long minutes= seconds/secondsPerMinute;
        seconds-=minutes*secondsPerMinute;
        StringBuffer sb = new StringBuffer();
        if(days>0){
            sb.append(days+"days");
        }
        if(hours<(5*secondsPerDay) && hours>0){
            if(sb.length()>0) sb.append(' ');
            sb.append(hours+"hrs");
        }
        if(minutes<(5*secondsPerHour) && minutes>0){
            if(sb.length()>0) sb.append(' ');
            sb.append(minutes+"mins");
        }
        if(secs<(5*secondsPerMinute) && seconds>0){
            if(sb.length()>0) sb.append(' ');
            sb.append(seconds+"s");
        }
        return sb.toString();
    }

    private String formatMetres(double metres) {
        if(metres<500){//0m-499m
           return String.valueOf(metres)+"m";
        }
        else if(metres>=500 && metres<100000){ //0.5km-99.9km
            BigDecimal bd = new BigDecimal(metres/1000f);
            bd=bd.setScale(1,BigDecimal.ROUND_HALF_UP);
            return String.valueOf(bd.floatValue())+"km";
        }
        else{ //100km-120km
            return String.valueOf((int)(metres/1000))+"km";
        }
    }


    private int findMidPoint(int x,int x1){
        if(x>x1) return (int) (((x-x1)/2f)+x1);
        else return (int) (((x1-x)/2f)+x);
    }
    private Point findMidPoint(Point p, Point p1) {
        Point p2=new Point(findMidPoint(p.x,p1.x),findMidPoint(p.y,p1.y));
        return p2;
    }

    /**
     *
     * @param pointMetres
     * @param boxMetres bounding box in metres, origin is bottom left of map
     * @return
     */
    private Point getPointPixels(Point pointMetres, Rectangle boxMetres) {
        int xMetresOffset=pointMetres.x-boxMetres.x;
        int xPixels = metresToPixels(xMetresOffset);
        int yMetresUpset=pointMetres.y - boxMetres.y;
        int yMetresDownset =  boxMetres.height-yMetresUpset;
        int yPixels=metresToPixels(yMetresDownset);
        return new Point(xPixels,yPixels);
    }

    ImageOperations io = new ImageOperations();
    private int metresToPixels(int metres){
        return metres/5;
    }
    private int pixelsToMetres(int pixels) {
        return pixels*5;
    }

    private OsGrid grid = new OsGrid();

    static class BadReferenceException extends Exception{
        String ref;
        public BadReferenceException(String ref, String message) {
            super(message);
            this.ref = ref;
        }

        public String getReference() {
            return ref;
        }
    }


    static class InvalidSquareRefException extends Exception{
        private String reference;
        public InvalidSquareRefException(Throwable cause, String reference) {
            super(cause);
            this.reference = reference;
        }
        public String getReference() {
            return reference;
        }
    }
    static class CannotGetSquareException extends Exception{
        private String reference;
        public CannotGetSquareException(Throwable cause, String reference) {
            super(cause);
            this.reference = reference;
        }

        public String getReference() {
            return reference;
        }
    }

}
