package uk.co.hcsoftware.images.tools;

import com.sun.media.jai.codec.FileSeekableStream;
import ns.TaskTimer;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;
import org.jdom.output.Format;
import uk.co.hcsoftware.yago.ArgReader;
import uk.co.hcsoftware.yago.Argument;
import uk.co.hcsoftware.yago.InvalidArgException;
import uk.co.hcsoftware.yago.Option;
import uk.co.hcsoftware.yago.parsers.GnuParser;
import uk.co.hcsoftware.yago.util.DefaultArgReaders;

import javax.media.jai.Interpolation;
import javax.media.jai.JAI;
import javax.media.jai.ParameterBlockJAI;
import javax.media.jai.PlanarImage;
import java.awt.image.renderable.ParameterBlock;
import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.FileOutputStream;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;


/**
 * scales a file to a particular output size (specify maxW, maxH) optionally can preserve aspect ratio
 *
 * create a toplevel file holding imagedetails e.g. <image path="galapagos/EPSN100.jpg" width="xx" height="yy"/>
 * then have gallery read the settings (if available) into the image class at parse-time
 *
 */
public class ThumbGenerator {
    private static final Logger log = Logger.getLogger(ThumbGenerator.class.getName());

    private static ArgReader<Integer> intReader = DefaultArgReaders.createIntReader();
    public static final Option<Integer> thumbMaxW = new Option<Integer>("t",true,intReader,"max width of thumbnail images");
    //new Option("tw","thumb_max_w",true,"(required) max width of thumbnail images");
    public static final Option<Integer> thumbMaxH = new Option<Integer>("th",false,intReader,"(defaults to t value) max height of thumbnail images");
    //Option("th","thumb_max_h",true,"(defaults to tw value) max height of thumbnail images");
    public static final Option<Integer> fullMaxW = new Option<Integer>("f",true,intReader,"max width of fullsize images");
    //new Option("fw","full_max_w",true,"(required) max width of fullsize images");
    public static final Option<Integer> fullMaxH = new Option<Integer>("fh",false,intReader,"(defaults to f value) max height of fullsize images");
    //new Option("fh","full_max_w",true,"(defaults to fw value) max height of fullsize images");

    private static ArgReader<File> dirReader=new ArgReader<File>() {
        public File readArg(String s) throws InvalidArgException {
            File f = new File(s);
            if(!f.exists()) throw new InvalidArgException("input dir doesn't exist!",s);
            if(!f.isDirectory()) throw new InvalidArgException("input dir isn't a directory!",s);
            return f;
        }
    };
    public static final Argument<File> inputDir = new Argument<File>(true,dirReader,"root_dir","dir from which to start processing");
    //public static final Option outputDir = new Option("o","output_dir",true,"(optional, default=inputdir) dir to which to copy ");

    private static ArgReader<String> stringReader = DefaultArgReaders.createStringReader();
    public static final Option<String> outputFormat = new Option<String>("outputformat", false,'o',"outputformat",stringReader,"format for output images (default=jpeg) see javax.media.jai.operator.FileStoreDescriptor for supported formats");

    public static final Option ignoreAspect = new Option("ignore aspect",false,'a',"ignoreaspect",null,"ignore aspect ratio (default keeps aspect ratio, so the longest side = maxvalue and the other side is scaled by the same amount)");
    public static final Option overwrite = new Option("overwrite",false,'w',"overwrite",null,"overwrite existing thumbs/fullsize images (default ignores the file if either a thumb or fullsize image already exists)");
    public static final Option<String> suffix = new Option<String>("suffix",false,'s',"suffix",stringReader,"suffix. Process only files having this suffix (ignores case) (default attempts to process all files)");
    public static final Option generateImagexml= new Option("imagexml",false,"generate image.xml file in root dir containing image details");
    private final ImageOperations imageOperations = new ImageOperations();

    public static void main(String[] argv) {
        try {
            Option[] opts = {thumbMaxW,thumbMaxH,fullMaxW,fullMaxH,outputFormat,ignoreAspect,overwrite,suffix,generateImagexml};
            Argument[] args={inputDir};
            new GnuParser(opts,args).parseOrDie("ThumbGenerator",argv
                    ,"Starts at the root_dir. recursively searches for imagefiles.\n" +
                    "For each image file, creates 2 directories at the same level: 'thumbs'" +
                    " and 'fullsize'.\n" +
                    "Creates a scaled version of the imagefile in each, based on option settings." +
                    "Also generates root_dir/images.xml containing information about the images");


            final File rootDir = inputDir.getArgValue();
            new ThumbGenerator(rootDir);
        } catch (IOException e) {
            log.log(Level.WARNING, "exception", e);
        }
    }
    Document imageFilesXml;
    private File imageRoot;
    public ThumbGenerator(File rootDir) throws IOException {
        this.imageRoot=rootDir;
        File f = new File(rootDir,"images.xml");
        if(generateImagexml.isSet()){
            if (!f.exists()) {
                imageFilesXml = new Document(new Element("images"));
            } else {
                try {
                    imageFilesXml = new SAXBuilder().build(f);
                } catch (JDOMException e) {
                    log.log(Level.WARNING,"e",e);
                    imageFilesXml = new Document(new Element("images"));
                }
            }
        }
        TaskTimer tt = new TaskTimer();
        process(rootDir);
        log.info("Completed. Took "+tt.stopString()+" to process "+filesProcessed+" image files");
        //save images.xml
        if(generateImagexml.isSet()){
            //File f = new File(rootDir,"images.xml");
            XMLOutputter xmlo = new XMLOutputter(Format.getPrettyFormat());
            final FileOutputStream fos = new FileOutputStream(f);
            xmlo.output(imageFilesXml,fos);
            fos.close();
        }
    }

    private static int filesProcessed=0;
    /**
     * process this directory and recurse into its subdirectories
     * @param inFile
     * @throws IOException
     */
    private void process(File inFile) throws IOException {
        //recursively look for jpegs in inFile (excluding dirs called thumbs or fullsize) and convert to thumbs
        //in a thumbs dir or fullsize images in a fullsize dir
        final List<String> excludedNames = Arrays.asList(new String[]{"thumbs","fullsize"});
        final String suffixVal = suffix.isSet()?suffix.getArgValue().toLowerCase() :null;


        File[] fs = inFile.listFiles(new FilenameFilter() {
            public boolean accept(File dir, String name) {
                File testFile=new File(dir,name);
                if(excludedNames.contains(name)) return false;
                else if(suffix.isSet() && !testFile.isDirectory()){
                    boolean b= name.toLowerCase().endsWith(suffixVal);
                    log.fine("returning "+b+" for file="+name);
                    return b;
                }
                return true;
            }
        });
        for(File imageFile:fs){
            if(imageFile.isDirectory()){
                process(imageFile);
            }
            else{
                File targetThumbFile=getTargetFile(imageFile,"thumbs");
                File targetFullsizeFile=getTargetFile(imageFile,"fullsize");

                if (!overwrite.isSet()) {
                    if(targetThumbFile.exists()){
                        log.info("ignoring "+imageFile+" as thumb="+targetThumbFile+" exists");
                        continue;
                    }
                    else if(targetFullsizeFile.exists()){
                        log.info("ignoring "+imageFile+" as fullsize="+targetFullsizeFile+" exists");
                        continue;
                    }
                }

                log.log(Level.INFO,"processing file="+imageFile);
                //generate thumb
                final int maxW = thumbMaxW.getArgValue().intValue();
                final int maxH = thumbMaxH.isSet()?thumbMaxH.getArgValue().intValue():maxW;
                PlanarImage thumb = genScaledImage(
                        targetThumbFile, imageFile
                        , maxW// Integer.parseInt(getValue(thumbMaxW))
                        , maxH// Integer.parseInt(getValue(thumbMaxH))
                );

                //generate fullsize
                final int fmaxW = fullMaxW.getArgValue().intValue();
                final int fMaxH = fullMaxH.isSet()?fullMaxH.getArgValue().intValue():fmaxW;
                PlanarImage fullsize = genScaledImage(
                        targetFullsizeFile ,imageFile
                        , fmaxW//  Integer.parseInt(getValue(fullMaxW))
                        , fMaxH// Integer.parseInt(getValue(fullMaxH))
                );
                if(generateImagexml.isSet()){//write imagedetails to file
                    Element root=imageFilesXml.getRootElement();
                    Element thumbElement = getImageElement(targetThumbFile,thumb);
                    Element fsElement = getImageElement(targetFullsizeFile,fullsize);
                    root.addContent(thumbElement);
                    root.addContent(fsElement);
                }
                filesProcessed++;
            }
        }

    }

    private Element getImageElement(File f,PlanarImage pi) {
        Element el = new Element("image");
        el.setAttribute("path",getRelativePath(imageRoot,f));
        el.setAttribute("width",String.valueOf(pi.getWidth()));
        el.setAttribute("height",String.valueOf(pi.getHeight()));
        return el;
    }

    private String getRelativePath(File imageRoot, File f) {
        String s=imageRoot.getAbsolutePath();
        String s1=f.getAbsolutePath();
        String s2= s1.substring(s.length()+1);
        if(s2.startsWith(File.separator)){ //strip filesep if there
            s2=s2.substring(1);
        }
        return s2;
    }

    private PlanarImage genScaledImage(File outFile, File imageFile, final int maxW, final int maxH) throws IOException {
       // File outFile = getTargetFile(imageFile, dir);
        return generateScaledImage(imageFile,outFile.getCanonicalPath() ,maxW ,maxH );
    }

    private File getTargetFile(File imageFile, final String dir) {
        File thumbsDir = new File(imageFile.getParentFile(),dir);
        thumbsDir.mkdirs();
        File outFile = new File(thumbsDir,imageFile.getName());
        return outFile;
    }

    private PlanarImage generateScaledImage(File inFile, String outFile, int maxW,int maxH) {
        /*
        * Create an input stream from the specified file name
        * to be used with the file decoding operator.
        */
        FileSeekableStream stream = null;
        try {
            stream = new FileSeekableStream(inFile);
        } catch (IOException e) {
            e.printStackTrace();
            System.exit(0);
        }
        final TaskTimer tt = new TaskTimer();

        /* Create an operator to decode the images file. */
        PlanarImage pi = JAI.create("stream", stream);
        final boolean preserveAspectRatio = !ignoreAspect.isSet();
        pi= imageOperations.scale(pi, maxW, maxH, preserveAspectRatio,Interpolation.INTERP_BICUBIC);
//        //do unsharp mask to smooth the result
//      //  pi=unsharpMask(pi,-0.33f);
//        final float amount = 1F;
//        pi = unsharpMask(pi, amount);

        ParameterBlock pb = new ParameterBlock();
        pb.addSource(pi);
        pb.add(outFile);
        final String of = outputFormat.isSet()?outputFormat.getArgValue():"jpeg";
        pb.add(of);
        pb.add(null);
        JAI.create("filestore", pb, null);
        log.log(Level.INFO,"wrote file= " + outFile+ " took "+tt.stopString());
        return pi;
    }

    private PlanarImage unsharpMask(PlanarImage pi, final float amount) {
        ParameterBlockJAI pbu = new ParameterBlockJAI("UnsharpMask");
        pbu.addSource(pi);

        pbu.setParameter("gain", amount);
        pi = JAI.create("UnsharpMask", pbu
                , null);
        return pi;
    }

}
