package uk.co.hcsoftware.images.web;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.Collections;
import java.util.Comparator;
import java.util.logging.Logger;
import java.util.zip.CRC32;


public class Gallery {
    final static Logger log = Logger.getLogger(Gallery.class.getName());

    private final String id,name;
    private static final String BLURBFILE_NAME = ".blurb";
    BlurbReader br;

    /**
     * default implentation uses the filesystem. ImageTree looks like this
     * rootDir
     *   gallery1
     *     thumbs
     *       thumb1
     *       thumb2
     *     fullsize
     *       fullsize1
     *       fullsize2
     *     original1
     *     original2
     *    gallery2
     *      ..
     *
     *
     *      */
    public Gallery(String id, File galleryDirectory, ImagesMetadata imd){
        this.id=id;
        this.name=galleryDirectory.getName();
        final File thumbDir = new File(galleryDirectory,"thumbs");
        final File fullsizeDir = new File(galleryDirectory,"fullsize");
        File[] files = galleryDirectory.listFiles();
        for(File f:files){
            if(f.isFile() && !f.getName().endsWith(BLURBFILE_NAME)){  //ignore blurbfiles
                File thumbFile = new File(thumbDir,f.getName());//get the thumbFile file
                File fullsizeFile = new File(fullsizeDir,f.getName());
                if(!thumbFile.exists()) log.warning("file "+thumbFile+" does not exist, ignoring file="+f);
                else if(!fullsizeFile.exists()) log.warning("file "+fullsizeFile+" does not exist, ignoring file="+f);
                else{
                    //get an ID for the image; use a checksum the galleryname+imagename (this should be unique)
                    //this helps with:
                    //1. if the name changes, or we move the image external links will still point to the correct image
                    //2. the ID doesn't reveal any extra info about the image itself (such as when it was taken, what camera etc)
                    //though there
                    CRC32 crc = new CRC32();
                    String unencoded = this.name+File.separatorChar+f.getName();
                    crc.update(unencoded.getBytes());
                    images.add(new Image(String.valueOf(crc.getValue()),f,thumbFile,fullsizeFile, imd));
                }
            }
        }
        //do we have some blurb?
        br = new BlurbReader(new File(galleryDirectory,BLURBFILE_NAME));
        
        //sort the images by filename
        Collections.sort(images,new Comparator<Image>() {
            public int compare(Image image, Image image1) {
                return image.getOrig().getName().compareTo(image1.getOrig().getName());
            }
        });
    }



    private final List<Image> images = new ArrayList<Image>();

    public List<Image> getImages(){
       return images;
    }
    public Image getImage(final String id) throws NoSuchImageException{
        for(Image i:images){
            if(id.equals(i.getId())){
                return i;
            }
        }
        throw new NoSuchImageException(id);
    }
    public Image getImage(final int number) throws NoSuchImageException{
        try {
            return images.get(number);
        } catch (IndexOutOfBoundsException e) {
             throw new NoSuchImageException(id);
        }
    }
    public int getImageCount(){
        return images.size();
    }

    public String getId() {
        return id;
    }

    public String getName() {
        return name;
    }
    public boolean hasBlurb(){
        return br.blurbExists();
    }
    public String getBlurb(){
        return br.getBlurb();
    }
}
